/*
 * IBM Confidential
 * OCO Source Materials
 * Component IBM Rational ROSE_RT BUILD : ROSE_RT
 * (c) Copyright IBM Corp. [1998], [2004] .All Rights Reserved.
 * 
 * The source code for this program is not published or
 * or otherwise divested of its trade secrets, irrespective
 * of what has been deposited with the U.S. Copyright Office.
 * 
 * 
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 * 
 */

#include <RTActor.h>
#include <RTActorRef.h>
#include <RTActor_class.h>
#include <RTComponentDescriptor.h>
#include <RTPortDescriptor.h>
#include <RTProtocol.h>
#include <RTTimerActor.h>
#include <RTTiming.h>

void RTActor::_predestroy( void )
{
    const RTActor_class    * info  = getActorData();
    RTTimerActor           * timer = RTTimerActor::soleInstance();
    const RTPortDescriptor * port;
    int                      count;
	int					     i;
    RTProtocol             * end;

    // disable binding notifications: we will soon purge any rtUnbound messages
    for( port = info->ports, count = info->numPorts; --count >= 0; ++port )
    {
	if( ( port->properties & RTPortDescriptor::MaskKind )
			      != RTPortDescriptor::KindSpecial )
	{
	    end = RTObjectOf( RTProtocol, port->offset, this );

	    end->bindingNotification( 0 );
	}
    }

    // destroy or deport all capsule roles
    const RTComponentDescriptor * component;

    count     = info->numComponents;
    component = &info->components[ 0 ];

    i = 0;

    while( i++ < count )
    {
		RTActorRef * ref = RTObjectOf( RTActorRef, component->offset, this );

		if( ref->isImported() != 0 )
			ref->deport();
		else
			(void)ref->destroy();

		component++;
    }

    // deregister all SAPs and SPPs, cancel any outstanding timer requests
    for( port = info->ports, count = info->numPorts; --count >= 0; ++port )
    {
	end = RTObjectOf( RTProtocol, port->offset, this );

	switch( port->properties & RTPortDescriptor::MaskKind )
	{
	case RTPortDescriptor::KindUnwired:
	    if( end->isRegistered() == 0 )
		break;

	    if( ( end->getFlags() & RTProtocol::RegisteredAsSAP ) != 0 )
		(void)end->deregisterSAP( RTProtocol::Automatic );
	    else
		(void)end->deregisterSPP( RTProtocol::Automatic );
	    break;

	case RTPortDescriptor::KindSpecial:
	    if( port->protocol == &Timing::Base::rt_class )
	    {
		// cancel all outstanding timer requests
		if( timer != (RTTimerActor *)0 )
		    timer->cancelAll( (Timing::Base *)end );
	    }
	    break;

	case RTPortDescriptor::KindWired:
	default:
	    break;
	}
    }
}
