/*
 * IBM Confidential
 * OCO Source Materials
 * Component IBM Rational ROSE_RT BUILD : ROSE_RT
 * (c) Copyright IBM Corp. [1998], [2004] .All Rights Reserved.
 * 
 * The source code for this program is not published or
 * or otherwise divested of its trade secrets, irrespective
 * of what has been deposited with the U.S. Copyright Office.
 * 
 * 
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 * 
 */


#ifndef __RTPubl_Message_h__
#define __RTPubl_Message_h__ included

#include <RTPubl/Priority.h>
#include <RTPubl/ProtDesc.h>
#include <RTPubl/Timespec.h>

struct _RTCapsule;
#if SEND_BY_VALUE
struct _RTObject_class;
#endif
struct _RTPort;

/*
    The following structures are not included as part of the publicly
    supported API, and thus should not be explicitly referenced in user
    code.
*/

typedef enum _RTMessage_Flag
{
    RTMessage_Flag_Deferred      = 0x01,
    RTMessage_Flag_Nondeferrable = 0x02,
    RTMessage_Flag_DataOwned     = 0x04
} RTMessage_Flag;

typedef enum _RTMessage_DeliverResult
{
    RTMessage_Accepted,
    RTMessage_Deferred,
    RTMessage_Rejected
} RTMessage_DeliverResult;

typedef struct _RTMessage__priv
{
    const void                   * userData;
    const struct _RTPort         * port;
    	
    RTSignal                       signal;
    RTPriority                     priority;

    RTFlags                        flags;
    RTPortIndex                    portIndex;

#if SEND_BY_VALUE
    const struct _RTObject_class * type;
#if RTMESSAGE_PAYLOAD_SIZE != 0
    /* space for a small amount of data */
    union
    {
	char                _data_area[ RTMESSAGE_PAYLOAD_SIZE ];
	RTAlignment         _alignment; /* to align data area */
    } data_union;
#endif /* RTMESSAGE_PAYLOAD_SIZE != 0 */
#endif /* SEND_BY_VALUE */

} RTMessage__priv;

typedef struct _RTMessage
{
    RTMessage__priv     _priv;

    struct _RTMessage * Next;
    struct _RTCapsule * capsule;

#if OBSERVABLE
    int                 _sendProbeId;
#endif

#if OTRTSDEBUG != DEBUG_NONE
	RTTimespec          _sendTime;
    RTPortIndex         _sendPortId;
    RTPortIndex         _sendPortIndex;
#endif /* OTRTSDEBUG != DEBUG_NONE */

} RTMessage;

/*
    Publicly supported interface for RTMessage
    Only the following methods are meant to be used in user code.
*/

/* defer this message, to be recalled later through RTPort_recall(At) */
#if MESSAGE_DEFERRAL
int RTMessage_defer( const RTMessage * );
#endif

#if SEND_BY_VALUE
int  RTMessage_copyData( const RTMessage *, void *, int );
#endif

RTS_INLINE RTSignal         RTMessage_getSignal   ( const RTMessage * );
#if RTS_NAMES
const char *     RTMessage_getSignalName( const RTMessage * );
#endif
RTS_INLINE const void     * RTMessage_getData     ( const RTMessage * );
RTS_INLINE RTPriority       RTMessage_getPriority ( const RTMessage * );
struct _RTPort * RTMessage_getPort     ( const RTMessage * );
int              RTMessage_getPortIndex( const RTMessage * );
#if SEND_BY_VALUE
RTS_INLINE const struct _RTObject_class * RTMessage_getType( const RTMessage * );
#endif

#if MESSAGE_DEFERRAL
RTS_INLINE    void   RTMessage_setDeferred( RTMessage * );
RTS_INLINE    void   RTMessage_resetDeferred( RTMessage * );
RTS_INLINE    int    RTMessage_isDeferred( const RTMessage * );
#endif

/*
    ------------------------------------------------
    private operations and structures
    The following methods and structures are not part of the
    supported API for use in user code.
*/

/* static (class) data */
typedef struct
{
    RTMessage iniMsg;
} RTMessage_Static_Type;

extern const RTMessage_Static_Type RTMessage_Static_Data;

/* allocate, deallocate, ctor, dtor */
RTMessage * new_RTMessage( void );
void        delete_RTMessage( RTMessage * );
void        RTMessage_construct( RTMessage * );

RTMessage_DeliverResult RTMessage_deliver( RTMessage * this );

void RTMessage_fill( RTMessage            *,
		     struct _RTCapsule    *,
		     const struct _RTPort *,
		     RTPortIndex,
		     RTSignal,
		     RTPriority,
		     const void           *,
		     RTFlags );

void RTMessage_copy ( RTMessage * this, const RTMessage * src );
void RTMessage_clear( RTMessage * this );

#if SEND_BY_VALUE
void RTMessage_setType( RTMessage *, const struct _RTObject_class * type );
#endif

#if RTS_INLINES
#include <RTPubl/Message.inl>
#endif

#endif /* __RTPubl_Message_h__ */
