#! perl -w

# IBM Confidential
# OCO Source Materials
# Component IBM Rational ROSE_RT BUILD : ROSE_RT
# (c) Copyright IBM Corp. [1998], [2004] .All Rights Reserved.
# 
# The source code for this program is not published or
# or otherwise divested of its trade secrets, irrespective
# of what has been deposited with the U.S. Copyright Office.
# 
# 
# Copyright (c) 1991-2003 Rational Software Corporation. All rights
# reserved.
# 
# Rational, the Rational logo, ObjecTime (and logo) are registered
# trademarks of Rational Software Corporation.
# 
# The placing of the copyright notices on Rational source code is not
# intended to mean that they are published works, but are included in the
# event of inadvertent publication. The source code of Rational contains
# trade secrets and other valuable confidential information of Rational
# Software Corporation. Rational Software Corporation and its licensors
# retain ownership to Rational Products in both source and object code
# format, and related documentation and all rights therein including
# copyrights, patents and trade secrets.
# 
# 
# U.S. Government Rights
# 
# Use, duplication, or disclosure by the U.S. Government is subject to
# restrictions set forth in the applicable Rational License Agreement
# and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
# 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
# or FAR 52.227-14, as applicable.
# 
# Unless otherwise agreed to in writing, use of Rational products is
# governed by the Rational Product License Agreement provided with
# the product.
# 

# Win9x command.com does not interact with perl according to win32 standard.
# We therefore need to call some functions in a non-standard way.
#
# This detects if the system is Win95 for that purpose.

sub mySystem
{
    my( $OS, $ROSERT_HOST );

    $OS          = '?' unless $OS          = $ENV{'OS'};
    $ROSERT_HOST = '?' unless $ROSERT_HOST = $ENV{'ROSERT_HOST'};

    if( $ROSERT_HOST eq 'win32' && $OS !~ m{^Windows_NT$}i )
    {
	return system( join( ' ', @_ ) );
    }
    else
    {
	return system( @_ );
    }
}

$debug   = 0;
$flat    = 0;
$out_sfx = '.c';

# Get the last-modified time of a file.

sub get_mtime
{
    my( $filename ) = @_;
    my( @info );

    if( @info = stat( $filename ) )
    {
	return $info[9];
    }
    else
    {
	return 0;
    }
}

# Filter duplicates out of the given list.

sub unique
{
    my( %seen ) = ();

    return grep { ++$seen{$_} == 1; } @_;
}

# Return a list of paths to the given list of files.

%cycles  = ();
%deplist = ();
%is_open = ();
%where   = ();

sub whereIs
{
    my( @result, $file );

    @result = ();
    foreach $file ( @_ )
    {
	push @result, "$where{$file}/$file";
    }

    return @result;
}

# Return a list of files included by the first file.

sub includesOf
{
    my( $depth, $file, @locations ) = @_;
    my( $FILE, $h, $loc, $inc, $level, @incs );

    if( ! defined( $where{$file} ) )
    {
	@{$deplist{$file}} = ();
	$where{$file}      = '';

	$FILE = "H$depth";

	foreach $loc ( @locations )
	{
	    next unless open( $FILE, "$loc/$file" );

	    $is_open{$file} = $depth;
	    $file{$depth}   = $file;

	    $where{$file} = $loc;

	    @incs = ();

	    while( <$FILE> )
	    {
		chomp;
		next unless s{^#\s*include\s*<([^>]*)>.*$}{$1}o;

		$inc = $_;

		if( $level = $is_open{$inc} )
		{
		    push @incs, $inc;
		    while( ++$level <= $depth )
		    {
			push @{$cycles{$inc}}, $file{$level};
		    }
		}
		else
		{
		    foreach $h ( &includesOf( $depth + 1, $inc, @h_locations ) )
		    {
			push @incs, $h;
		    }
		    push @incs, $inc if $where{$inc} ne '';
		}
	    }

	    undef $is_open{$file};
	    close( $FILE );

	    @{$deplist{$file}} = sort( &unique( @incs ) );
	    last;
	}

	while( $inc = pop( @{$cycles{$file}} ) )
	{
	    @{$deplist{$inc}} = sort( &unique( @{$deplist{$inc}},
					       @{$deplist{$file}} ) );
	}
    }

    return @{$deplist{$file}};
}

# Read the configuration file and make sure all the pieces are available.

sub read_config
{
    my( $config ) = @_;
    my( $target, $libset );
    my( @required, @missing );

    if( $config =~ m{^([^.]*)([ST])\.(.*)$} )
    {
	$target_base =  $1;
	$target      = "$1$2";
	$libset      =  $3;
    }
    else
    {
	print STDERR "Malformed configuration name: '$config'\n";
	exit 1;
    }

    $build_dir = "build-$config";

    $include_opt   = '-I';
    @make_args     = ( '-f', '../src/main.mk' );
    $object_ext    = '.o';
    $preprocessor  = undef;
    $postprocessor = undef;
    $supported     = 'No';

    if( ! -e "config/$config/setup.pl" )
    {
	print STDERR "Can't find 'config/$config/setup.pl'\n";
	exit 1;
    }

    if( ! do "config/$config/setup.pl" )
    {
	print STDERR "config/$config/setup.pl: $!\n";
	exit 1;
    }

    @source_path = split( ',', $target_base );

    @required = (
		    "config/$config/config.mk",
		    "libset/$libset/libset.mk",
		    "target/$target/target.mk",
		    "target/$target/RTTarget.h",
		    ( map "src/target/$_/.", @source_path )
		);

    @missing = ();
    foreach $path ( @required )
    {
	push( @missing, $path ) unless -e $path;
    }

    if( $#missing >= 0 )
    {
	printf STDERR "The following files and directories are required\n"
		    . "by the configuration '$config':\n\t%s\n",
		      join( "\n\t", @missing );
	exit 1;
    }

    return ( $target, $libset, @source_path );
}

# Display configuration information.

sub show_info
{
    printf "*** $config%s ***\n\n",
	    ( $supported eq 'Yes' ? '' : ' (unsupported)' );
}

# Create a directory if it does not already exist.

sub make_dir
{
    my( $dir ) = $_[0];

    if( ! -d $dir && ! mkdir( $dir, 0777 ) )
    {
	print STDERR "$dir: $!\n";
	exit 1;
    }
}

# Create all necessary directories.

sub make_directories
{
    &make_dir( 'lib' );
    &make_dir( "lib/$config" );
    &make_dir( $build_dir );
    if( $flat )
    {
	&make_dir( "$build_dir/RTPubl" );
	&make_dir( "$build_dir/RTPriv" );
    }
}

sub show_words
{
    my( @words ) = @_;
    my( $word, $word_width, $line_width );

    $line_width = 78;
    foreach $word ( @words )
    {
	$word_width  = length( $word ) + 1;
	$line_width += $word_width;
	if( $line_width > 78 )
	{
	    print DEPEND " \\\n";
	    $line_width = $word_width;
	}
	print DEPEND ' ', $word;
    }
    print DEPEND "\n";
}

package SourceFile;

$changed    = 0;
$in_sfx     = '.c';

$nextObjId  = 'aaa';
%objId      = ();
$missing    = 0;

# These must be avoided when choosing object file names
# because they are special in some environments.
%badObjId   = ( 'aux' => 1, 'con' => 1  );

$nextFlatId = 'FlatAA';
%flatId     = ();
%group      = ();

sub load
{
    my( $filename ) = @_;

    print "Pre-loading from '$filename'\n" if $main::debug;

    if( ! open( DB, "<$filename" ) )
    {
	print STDERR "Can't read '$filename'.\n";
	exit 1;
    }

    while( <DB> )
    {
	chomp;
	if( m{^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$} )
	{
	    $obj    = $1;
	    $group  = $2;
	    $class  = $3;
	    $member = $4;
	}
	else
	{
	    print STDERR "Masformed entry: '$_'\n";
	    $changed = 1;
	    next;
	}

	$src = "$class/$member";

	if( defined( $group{$src} ) )
	{
	    $changed = 1;
	}
	else
	{
	    $group{$src} = $group;
	}

	if( defined( $objId{$src} ) )
	{
	    print STDERR "Duplicate entry: '$_'\n";
	    $changed = 1;
	    next;
	}

	$objId{$src} = $obj;
	++$missing;

	$nextObjId = $obj if $nextObjId lt $obj;
    }

    ++$nextObjId;

    close( DB );
}

sub store
{
    my( $filename, @sources ) = @_;
    my( $source );

    if( ! open( DB, ">$filename" ) )
    {
	print STDERR "Can't write to '$filename'.\n";
	exit 1;
    }

    binmode( DB );

    foreach $source ( @sources )
    {
	print DB $source->object, ' ',
		 $source->group,  ' ',
		 $source->class,  ' ',
		 $source->member, "\n";
    }

    close( DB );
}

sub new
{
    my( $type, $group, $class, $member ) = @_;

    my $this = bless {};

    $this->{'group'}  = $group;
    $this->{'class'}  = $class;
    $this->{'member'} = $member;

    $src = "$class/$member";

    if( ! defined( $group{$src} ) )
    {
	$group{$src} = $group;
	$changed = 1;
    }
    elsif( $group{$src} ne $group )
    {
	$changed = 1;
    }

    if( ! defined( $obj = $objId{$src} ) )
    {
	$changed = 1;
    }

    --$missing;

    if( $group eq 'M' )
    {
	$obj = "\$(RTS_LIBRARY)/$member";
    }
    elsif( ! defined( $obj ) )
    {
	do
	{
	    $obj = $nextObjId++;
	}
	while( defined( $badObjId{$obj} ) );
    }

    if( $main::flat )
    {
	if( ! defined( $group{$class} ) )
	{
	    $group{$class} = $group;
	}

	if( $group eq 'M' )
	{
	    $flatSrc = "${class}_$member";
	}
	else
	{
	    $flatSrc  =   $class;
	    $flatSrc .= "_$group" if $group ne $group{$class};

	    $obj = $flatSrc;
	}

	$objId{$flatSrc} = $obj;

	if( ! defined( $flatId{$flatSrc} ) )
	{
	    @{$nextFlatId} = ();
	    $flatId{$flatSrc} = $nextFlatId++;
	}

	push @{$flatId{$flatSrc}}, $this;
    }

    $this->{'object'} = $objId{$src} = $obj;

    return $this;
}

sub class  { return shift->{'class'};  }
sub group  { return shift->{'group'};  }
sub member { return shift->{'member'}; }
sub object { return shift->{'object'}; }

sub source
{
    my( $this ) = shift;

    return "$this->{'class'}/$this->{'member'}$in_sfx";
}

sub flat_keys
{
    return keys %flatId;
}

sub flat_object
{
    my( $class ) = @_;

    return @{$flatId{$class}}[0]->object;
}

sub flat_sources
{
    my( $class ) = @_;

    return map( { $_->source } @{$flatId{$class}} );
}

sub no_changes
{
    my( $class ) = @_;

    return $changed == 0 && $missing == 0;
}

package main;

# Make sure the dependency rules are up to date or recompute them.

sub make_depend
{
    $depend_mk    = 'depend.mk';
    $manifest_src = "../src/MANIFEST$out_sfx";
    $manifest_i   = 'MANIFEST.i';
    $object_map   = 'depend.dat';

    # The following pathnames are relative to the build directory.

    @h_public =
	(
	    "../build-$config",	
	    "../config/$config",
	    "../libset/$libset",
	    "../target/$target",
	    '../include'
	);

    @h_private = grep { -d $_ }
	(
		  "../src/config/$config",
		  "../src/libset/$libset",
	    ( map "../src/target/$_", @source_path ),
		  '../src/include'
	);

    @h_locations = ( @h_public, @h_private );

    @s_locations = grep { -d $_ }
	(
	    "../src/config/$config",
	    "../src/libset/$libset",
	    ( map "../src/target/$_", @source_path ),
	    '../src'
	);

    $preprocess = sprintf( "$preprocessor%s %s",
			   join( " $include_opt", '', @h_locations ),
			   $manifest_src );

    printf "Pre-processor: '%s'\n", $preprocess if $debug;

    if( system( $preprocess ) != 0 )
    {
	print STDERR "Pre-processor failed.\n";
	exit 1;
    }

    if( $postprocessor )
    {
	if( system( $postprocessor ) != 0 )
	{
	    print STDERR "Post-processor failed.\n";
	    exit 1;
	}
    }

    if( -e $object_map )
    {
	&SourceFile::load( $object_map );
    }

    %classes = ();
    @sources = ();

    $group       = undef;
    @groups      = ();
    $nextGroupId = 'GroupAA';
    %groupId     = ();

    @headers = ();

    if( ! open( MANIFEST, "<$manifest_i" ) )
    {
	print STDERR "Can't read '$manifest_i'.\n";
	exit 1;
    }

    $rc = 0;
    $in_comment = 0;

    while( <MANIFEST> )
    {
	# Remove the trailing new-line.
	chomp;

	if( $in_comment )
	{
	    if( s{.*?\*/}{} )
	    {
		$in_comment = 0;
	    }
	    else
	    {
		next;
	    }
	}

	s{/\*.*?\*/}{}g;
	s{//.*$}{};
	s{\s+$}{};

	if( s{/\*.*$}{} )
	{
	    $in_comment = 1;
	}

	if( m{^\s*#} )
	{
	    # Ignore '#line' directives.
	}
	elsif( m{^\s*GROUP\s+(\S+)$} )
	{
	    # Watch for library grouping.
	    $group = $1;

	    if( ! defined( $groupId{$group} ) )
	    {
		push @groups, $group;
		@{$nextGroupId}  = ();
		$groupId{$group} = $nextGroupId++;
	    }
	}
	elsif( ! $group )
	{
	    # Ignore stuff before the first group.
	}
	elsif( m{^\s*(\S+)\s+(\S+)$} )
	{
	    $class  = $1;
	    $member = $2;
	    push @sources, new SourceFile( $group, $class, $member );
	}
	elsif( m{^\s*(\S+)\s+(\S+)\s*(.*)$} )
	{
	    $class  = $1;
	    $member = $2;
	    if( eval( $3 ) )
	    {
		push @sources, new SourceFile( $group, $class, $member );
	    }
	}
	elsif( m{\s} )
	{
	    print STDERR "Improper input: '$_'\n";
	    $rc = 1;
	}
    }

    close( MANIFEST );

    unlink( $manifest_i );

    exit $rc if $rc != 0;

    if( &get_mtime( $depend_mk ) > &get_mtime( $manifest_src ) &&
	&SourceFile::no_changes() )
    {
	return; # up to date
    }

    print "Computing dependencies ...\n";

    if( ! open( DEPEND, ">$depend_mk" ) )
    {
	print STDERR "Can't write to '$depend_mk'.\n";
	exit 1;
    }

    binmode( DEPEND );

    print DEPEND "# This file was automatically generated;\n";
    print DEPEND "# you probably don't want to edit it manually.\n";

    print DEPEND "\nGLOBAL_DEPEND =";
    &show_words( "../config/$config/config.mk",
		 "../libset/$libset/libset.mk",
		 "../target/$target/target.mk" );

    print DEPEND "\nCC_CMD =";
    &show_words( '$(CC)',
		 '$(LIBSETCCFLAGS)', '$(TARGETCCFLAGS)', '$(LIBSETCCEXTRA)',
		 ( map "$include_opt$_", @h_locations ),
		 '$(OBJECT_OPT)', '$(OBJOUT_OPT)', '$(OBJOUT_TAG)$@' );

    @sources = sort { $a->source cmp $b->source } @sources;

    foreach $source ( @sources )
    {
	$group = $source->group;
	$class = $source->class;
	$obj   = $source->object;
	$src   = $source->source;

	push @{${class}.'.all'}, $obj;
	push @{$groupId{$group}}, $obj;

	$classes{$class} = 1;

	if( ! defined( $where{$src} ) )
	{
	    @prereq = &includesOf( 1, $src, @s_locations );

	    if( ! $where{$src} )
	    {
		print STDERR "Can't find '$src'\n";
		$rc = 1;
	    }
	    elsif( $flat )
	    {
		push @headers, @prereq;
	    }
	    else
	    {
		( $cc ) = &whereIs( $src );
		print DEPEND "\n$obj$object_ext :";
		&show_words( $cc, &whereIs( @prereq ), '$(GLOBAL_DEPEND)' );
		print DEPEND "\t\@\$(CC_ECHO) $src\n";
		print DEPEND "\t\@\$(CC_CMD) $cc\n";
	    }
	}
    }

    foreach $group ( @groups )
    {
	print DEPEND "\n${group}_OBJECTS=";
	&show_words( map( "$_$object_ext", &unique( @{$groupId{$group}} ) ) );
    }

    foreach $class ( sort &SourceFile::flat_keys )
    {
	my( @sources ) = &SourceFile::flat_sources( $class );

	print DEPEND "\n${class}_SRC=";
	&show_words( &whereIs( sort( &unique( @sources ) ) ) );
	print DEPEND "\n$class$out_sfx : \$(${class}_SRC)\n";
	print DEPEND "\t\@\$(MERGE) \$(${class}_SRC) >\$\@\n\n";

	print DEPEND &SourceFile::flat_object( $class ),
		     "$object_ext : $class$out_sfx";
	&show_words( sort( &unique( map( &includesOf( 1, $_ ), @sources ) ) ),
		     '$(GLOBAL_DEPEND)' );

	print DEPEND "\t\@\$(CC_ECHO) $class\n";
	print DEPEND "\t\@\$(CC_CMD) $class$out_sfx\n";
    }

    foreach $header ( sort( &unique( @headers ) ) )
    {
	print DEPEND "\n$header : ", &whereIs( $header ), "\n";
	print DEPEND "\t\@\$(MERGE) \$? >\$\@\n";
    }

    foreach $class ( sort keys %classes )
    {
	print DEPEND "\n${class}.all :";
	&show_words( map( "$_$object_ext", &unique( @{${class}.'.all'} ) ) );
	print DEPEND "\t\@\$(NOP)\n";
    }

    close( DEPEND );

    if( $rc != 0 )
    {
	unlink( $depend_mk );
	exit $rc;
    }

    foreach $group ( @groups )
    {
	if( ! open( OBJECTS, ">$group.olist" ) )
	{
	    print STDERR "Can't write to '$group.olist'.\n";
	    exit 1;
	}

	binmode( OBJECTS );

	foreach $obj ( &unique( @{$groupId{$group}} ) )
	{
	    print OBJECTS "$obj\n";
	}

	close( OBJECTS );
    }

    &SourceFile::store( $object_map, @sources );
}

# main body

# Don't buffer output.

$| = 1;

# Move up to the root of the RTS source tree.

chdir( '..' );

# Parse arguments

if( $#ARGV >= 0 && $ARGV[0] eq '-flat' )
{
    $flat = 1;
    shift;
}

if( $#ARGV < 0 )
{
    print STDERR "Usage: $0 [-flat] <config> [make options ...]\n";
    exit 1;
}

$config = $ARGV[0];
shift;

# Read configuration information.

( $target, $libset, @source_path ) = &read_config( $config );

# Display configuration information.

&show_info();

# Make sure the necessary directories exist.

&make_directories();

# Move to the build directory.

chdir( $build_dir );

# Make sure the dependency rules are up to date.

&make_depend();

if( $#ARGV < 0 )
{
    $rc = 0;
}
else
{
    # Use make to do the rest.

    @make_cmd = ( @ARGV,
		  "LIBRARY_SET=$libset",
		  "TARGET=$target",
		  @make_args
		);

    printf "%s\n", join( ' ', @make_cmd ) if $debug;

    $rc = &mySystem( @make_cmd );

    print "\n";
}

exit $rc;
